import os
import argparse
import logging
import subprocess
import shlex
import uuid
import threading
import multiprocessing as mp
import tempfile
from future.utils import string_types
from fanc.config import config
import re


# configure logging
logger = logging.getLogger(__name__)


class CommandTask(object):
    def __init__(self, command):
        if isinstance(command, string_types):
            self.command = shlex.split(command, posix=False)
        else:
            self.command = command
        self.id = uuid.uuid4()


class QueuedTask(object):
    def __init__(self, task, wait_for=None, threads=1):
        self.task = task
        self.id = task.id
        self.command = task.command

        if isinstance(wait_for, CommandTask):
            wait_for = [wait_for]
        elif wait_for is None:
            wait_for = []

        self.wait_for = wait_for
        self.threads = threads


class TaskRunner(object):
    def __init__(self):
        self._tasks = []
        self._task_ixs = dict()

    def add_task(self, task, wait_for=None, threads=1, *args, **kwargs):
        self._tasks.append(QueuedTask(task, wait_for=wait_for, threads=threads))
        self._task_ixs[task.id] = len(self._task_ixs)

    def run(self, *args, **kwargs):
        raise NotImplementedError("Subclasses of TaskWorker must implement 'run'")


def _run_task(task, process_id, completed_queue):
    res = subprocess.call(task.command)
    completed_queue.put((task, process_id, res))


class ParallelTaskRunner(TaskRunner):
    def __init__(self, threads=1, test=False):
        TaskRunner.__init__(self)
        self._threads = threads
        self._active_threads_lock = threading.Lock()
        self._active_threads = 0
        self._test = test

    def _change_active_threads(self, threads):
        with self._active_threads_lock:
            self._active_threads += threads

    def run(self, *args, **kwargs):
        completed = set()
        running = set()
        completed_queue = mp.Queue()
        processes = dict()

        remaining_tasks = len(self._tasks)
        while remaining_tasks > 0:

            # find tasks that can be executed
            for task in self._tasks:
                if task.id in completed:
                    continue

                if task.id in running:
                    continue

                if task.threads > self._threads:
                    raise RuntimeError("Task requests more threads ({}) than "
                                       "available in parallel task runner ({})!".format(task.threads,
                                                                                        self._threads))

                available_threads = self._threads - self._active_threads
                if task.threads > available_threads:
                    continue

                is_executable = True
                for wait_task in task.wait_for:
                    if not wait_task.id in completed:
                        is_executable = False
                        break

                if not is_executable:
                    continue

                p_id = uuid.uuid4()
                if not self._test:
                    logger.debug("Running task {}".format(task.id))
                    p = mp.Process(target=_run_task, args=(task, p_id, completed_queue))
                    p.start()
                    processes[p_id] = p
                else:
                    print("{} (threads: {}, depends on: {}): {}".format(
                        self._task_ixs[task.id],
                        task.threads,
                        ", ".join([str(self._task_ixs[t.id])
                                   for t in task.wait_for]),
                        " ".join(task.command))
                    )

                    completed_queue.put((task, p_id, 0))
                running.add(task.id)
                self._change_active_threads(task.threads)

            # wait for any task to finish
            task, p_id, res = completed_queue.get(block=True)
            completed.add(task.id)
            running.remove(task.id)
            self._change_active_threads(-task.threads)
            if not self._test:
                processes[p_id].join()
                del processes[p_id]

            if res != 0:
                for p in processes.values():
                    p.terminate()
                    p.join()
                raise RuntimeError("Task {} had non-zero exit status. "
                                   "Cancelling execution of all tasks.".format(task.id))
            else:
                remaining_tasks -= 1


class SgeTaskRunner(TaskRunner):
    def __init__(self, task_prefix=None,
                 log_dir=None, trap_sigusr=True,
                 startup_commands_file=None,
                 cleanup_commands_file=None,
                 buffer_python_output=False):
        TaskRunner.__init__(self)
        if task_prefix is None:
            from fanc.tools.files import random_name
            self._task_prefix = 'fanc_' + random_name(6) + '_'
        else:
            self._task_prefix = task_prefix

        self._log_dir = log_dir
        if log_dir is None:
            self._log_dir = config.sge_log_dir
        else:
            self._log_dir = os.path.expanduser(log_dir)
        self._trap_sigusr = trap_sigusr
        self._startup_commands_file = startup_commands_file
        self._cleanup_commands_file = cleanup_commands_file
        self._buffer_python_output = buffer_python_output

    def _submit_task(self, task, kill=True):
        task_ix = self._task_ixs[task.id]
        job_id = self._task_prefix + '{}'.format(task_ix)
        with tempfile.NamedTemporaryFile('w', prefix='fanc_auto_', suffix='.sh') as tmp_file:
            if self._trap_sigusr:
                tmp_file.write('function notify_handler() {\n  '
                               '(>&2 echo "Received termination notice")\n}\n')
                tmp_file.write("trap notify_handler SIGUSR1\n")
                tmp_file.write("trap notify_handler SIGUSR2\n\n")
            if not self._buffer_python_output:
                tmp_file.write("export PYTHONUNBUFFERED=1\n\n")

            if self._startup_commands_file is not None:
                with open(self._startup_commands_file) as f:
                    startup_commands = f.read()
                tmp_file.write(startup_commands)
                tmp_file.write("\n")

            tmp_file.write(" ".join(task.command) + "\n")
            tmp_file.write("OUT=$?\n")
            tmp_file.write("if [ $OUT -ne 0 ]; then\n")
            tmp_file.write('    (>&2 echo "Job $JOB_ID / $JOB_NAME had non-zero exit status")\n')
            if kill is not None:
                if self._log_dir is not None:
                    tmp_file.write(
                        '    echo "Failed job ID: {}" >> {}\n'.format(
                            job_id, os.path.join(self._log_dir, self._task_prefix + "FAILED")
                        ))
                tmp_file.write('    {} "{}*"\n'.format(config.sge_qdel_path, self._task_prefix))
            tmp_file.write('    res="failed"\n')
            tmp_file.write('else\n')
            tmp_file.write('    res="succeeded"\n')
            tmp_file.write("fi\n\n")

            if self._cleanup_commands_file is not None:
                with open(self._cleanup_commands_file) as f:
                    cleanup_commands = f.read()
                tmp_file.write(cleanup_commands)
                tmp_file.write("\n")

            tmp_file.flush()

            command = [config.sge_qsub_path,
                       '-N', job_id, '-cwd',
                       '-pe', config.sge_parallel_environment,
                       str(task.threads)] + config.sge_qsub_options.split()

            if config.sge_default_queue is not None:
                command += ['-q', config.sge_default_queue]

            if config.sge_shell is None:
                shell = '/bin/bash'
            else:
                shell = config.sge_shell
            command += ['-S', shell]

            if task.wait_for is not None and len(task.wait_for) > 0:
                hold_ids = ",".join([self._task_prefix + '{}'.format(self._task_ixs[t.id])
                                     for t in task.wait_for])
                command += ['-hold_jid', hold_ids]

            if self._log_dir is not None:
                command += ['-o', os.path.join(self._log_dir, job_id + '_o')]
                command += ['-e', os.path.join(self._log_dir, job_id + '_e')]
            else:
                command += ['-o', '/dev/null']
                command += ['-e', '/dev/null']

            command += [tmp_file.name]

            logger.info("Submitting {}".format(" ".join(command)))
            subprocess.call(command)

    def run(self, *args, **kwargs):
        for task in self._tasks:
            self._submit_task(task)


class SlurmTaskRunner(TaskRunner):
    def __init__(self, task_prefix=None,
                 log_dir=None, trap_sigusr=True,
                 startup_commands_file=None,
                 cleanup_commands_file=None,
                 buffer_python_output=False):
        TaskRunner.__init__(self)
        if task_prefix is None:
            from fanc.tools.files import random_name
            self._task_prefix = 'fanc_' + random_name(6) + '_'
        else:
            self._task_prefix = task_prefix

        self._log_dir = log_dir
        if log_dir is None:
            self._log_dir = config.slurm_log_dir
        else:
            self._log_dir = os.path.expanduser(log_dir)
        self._trap_sigusr = trap_sigusr

        self._slurm_job_ids = dict()
        self._startup_commands_file = startup_commands_file
        self._cleanup_commands_file = cleanup_commands_file
        self._buffer_python_output = buffer_python_output

    def _submit_batch_command(self, task, kill=True):
        job_name = self._task_prefix
        with tempfile.NamedTemporaryFile('w', prefix='fanc_auto_', suffix='.sh') as tmp_file:
            tmp_file.write("#!{}\n".format(config.slurm_shell if config.slurm_shell is not None else "/bin/bash"))
            tmp_file.write("#SBATCH --nodes=1\n")
            tmp_file.write("#SBATCH --ntasks=1\n")
            tmp_file.write("#SBATCH --cpus-per-task={}\n".format(task.threads))
            tmp_file.write("#SBATCH --job-name={}\n".format(job_name))
            if self._log_dir is not None:
                task_ix = self._task_ixs[task.id]
                log_prefix = self._task_prefix + str(task_ix)
                tmp_file.write("#SBATCH --output={}\n".format(os.path.join(self._log_dir, log_prefix + '_o')))
                tmp_file.write("#SBATCH --error={}\n".format(os.path.join(self._log_dir, log_prefix + '_e')))
            else:
                tmp_file.write("#SBATCH --output={}\n".format('/dev/null'))
                tmp_file.write("#SBATCH --error={}\n".format('/dev/null'))

            if task.wait_for is not None and len(task.wait_for) > 0:
                wait_job_ids = []
                for t in task.wait_for:
                    try:
                        wait_job_ids.append(self._slurm_job_ids[t.id])
                    except KeyError:
                        raise RuntimeError("Task {} has not been submitted to Slurm yet!".format(t.id))

                dependencies = "afterok:{}".format(":".join(wait_job_ids))
                tmp_file.write("#SBATCH --dependency={}\n".format(dependencies))

            tmp_file.write("#SBATCH --export=ALL\n")

            if not self._buffer_python_output:
                tmp_file.write("export PYTHONUNBUFFERED=1\n\n")

            if self._startup_commands_file is not None:
                with open(self._startup_commands_file) as f:
                    startup_commands = f.read()
                tmp_file.write(startup_commands)
                tmp_file.write("\n")

            if self._trap_sigusr:
                tmp_file.write('function notify_handler() {\n  '
                               '(>&2 echo "Received termination notice")\n}\n')
                tmp_file.write("trap notify_handler SIGUSR1\n")
                tmp_file.write("trap notify_handler SIGUSR2\n\n")

            tmp_file.write(" ".join(task.command) + "\n")
            tmp_file.write("OUT=$?\n")
            tmp_file.write("if [ $OUT -ne 0 ]; then\n")
            tmp_file.write('    (>&2 echo "Job $JOB_ID / $JOB_NAME had non-zero exit status")\n')
            if kill is not None:
                if self._log_dir is not None:
                    tmp_file.write(
                        '    echo "Failed job ID: {}" >> {}\n'.format(
                            job_name, os.path.join(self._log_dir,
                                                   self._task_prefix + "FAILED")
                        ))
                tmp_file.write('    {} --signal=USR1 -n "{}"\n'.format(config.slurm_scancel_path,
                                                                       self._task_prefix))
            tmp_file.write('    res="failed"\n')
            tmp_file.write('else\n')
            tmp_file.write('    res="succeeded"\n')
            tmp_file.write("fi\n\n")

            if self._cleanup_commands_file is not None:
                with open(self._cleanup_commands_file) as f:
                    cleanup_commands = f.read()
                tmp_file.write(cleanup_commands)
                tmp_file.write("\n")

            tmp_file.flush()

            with open(tmp_file.name, 'r') as f:
                for line in f:
                    line = line.rstrip()
                    print(line)

            command = [config.slurm_sbatch_path] + config.slurm_sbatch_options.split()

            command += [tmp_file.name]

            slurm_job_id = subprocess.check_output(command)
            slurm_job_id = slurm_job_id.rstrip()
            if isinstance(slurm_job_id, bytes):
                slurm_job_id = slurm_job_id.decode('utf-8')
            m = re.search(r'(\d+)', slurm_job_id)
            slurm_job_id = m.group(1)

            return slurm_job_id

    def run(self, *args, **kwargs):
        completed = set()

        remaining_tasks = len(self._tasks)
        while remaining_tasks > 0:

            # find tasks that can be executed
            for task in self._tasks:
                if task.id in completed:
                    continue

                is_executable = True
                for wait_task in task.wait_for:
                    if wait_task.id not in completed:
                        is_executable = False
                        break

                if not is_executable:
                    continue

                # all dependencies submitted
                self._slurm_job_ids[task.id] = self._submit_batch_command(task)
                completed.add(task.id)
                remaining_tasks -= 1


def auto_parser():
    parser = argparse.ArgumentParser(
        prog="fanc auto",
        description='Automatically process an entire Hi-C data set.'
    )

    parser.add_argument(
        'input',
        nargs='+',
        help="Input files. "
             "fanc will try to guess the file "
             "type by its extension."
    )

    parser.add_argument(
        'output_folder',
        help="Output folder. "
             "All output files and folders " 
             "will be generated under this directory."
    )

    parser.add_argument(
        '-g', '--genome', dest='genome',
        help="Genome for the Hi-C object." 
             "Path to region-based file (BED, GFF, ...) containing "
             "the non-overlapping regions to be used for Hi-C "
             "object construction. Typically restriction-enzyme fragments. "
             "Alternatively: Path to genome file (FASTA, folder with "
             "FASTA, hdf5 file), which will be used in conjunction "
             "with the type of restriction enzyme (-r) to calculate "
             "fragments directly."
    )

    parser.add_argument(
        '-r', '--restriction-enzyme', dest='restriction_enzyme',
        help="Restriction enzyme name. "
             "Used for in silico digestion "
             "of genomic sequences and splitting of reads at Hi-C "
             "ligation junctions. (e.g. HindIII, case-sensitive). "
             "Separate multiple enzymes with ','. "
             "Restriction names can be any supported by Biopython, which obtains data "
             "from REBASE (http://rebase.neb.com/rebase/rebase.html). "
    )

    parser.add_argument(
        '-i', '--genome-index', dest='genome_index',
        help="Bowtie 2 or BWA genome index. "
             "Only required when passing FASTQ "
             "files as input."
    )

    parser.add_argument(
        '-n', '--basename', dest='basename',
        help="Basename for output files. " 
             "If not provided, will be guessed based " 
             "on input file names."
    )

    parser.add_argument(
        '-s', '--step-size', dest='step_size',
        type=int,
        default=3,
        help="Step size for iterative mapping. " 
             "Default: %(default)d"
    )

    parser.add_argument(
        '-b', '--bin-sizes', dest='bin_sizes',
        nargs='+',
        default=['5mb', '2mb', '1mb', '500kb', '250kb', '100kb', '50kb',
                 '25kb', '10kb', '5kb'],
        help="Bin sizes for Hi-C matrix generation. " 
             "Default: 5mb, 2mb, 1mb, "
             "500kb, 250kb, 100kb, 50kb, 25kb, " 
             "10kb, 5kb."
    )

    parser.add_argument(
        '-t', '--threads', dest='threads',
        type=int,
        default=1,
        help="Maximum number of threads. "
             "The number provided here will not be exceeded " 
             "by analysis steps running alone or in parallel. "
             "Default: %(default)d"
    )

    parser.add_argument(
        '--fanc-parallel', dest='mapper_parallel',
        action='store_false',
        default=True,
        help='Use FAN-C parallelisation, which launches multiple mapper jobs. '
             'This may be faster in some cases than relying '
             'on the internal parallelisation of the mapper, '
             'but has potentially high disk I/O and memory usage.'
    )

    parser.add_argument(
        '--split-fastq', dest='split_fastq',
        action='store_true',
        default=False,
        help="Split fastq files into chunks of 10M reads. " 
             "Reads will be merged again on the SAM level. "
             "Splitting and merging bypasses the -tmp flag. " 
             "This option reduces disk usage in tmp, in case "
             "the system has a small tmp partition. "
    )

    parser.add_argument(
        '--memory-map', dest='memory_map',
        action='store_true',
        default=False,
        help="Map Bowtie2 index to memory. Recommended " 
             "if running on medium-memory systems and using many "
             "parallel threads)."
    )

    parser.add_argument(
        '--ice', dest='ice',
        action='store_true',
        default=False,
        help="Correct Hi-C matrices using ICE instead of Knight-Ruiz "
             "matrix balancing. Slower, but much more memory-friendly."
    )

    parser.add_argument(
        '-q', '--quality-cutoff', dest='quality_cutoff',
        type=float,
        help='Cutoff for the minimum mapping quality of a read. '
             'For numbers larger than 1, will filter on MAPQ. '
             'If a number between 0 and 1 is provided, will filter '
             'on the AS tag instead of mapping quality (only BWA). '
             'The quality cutoff is then interpreted as the '
             'fraction of bases that have to be matched for any '
             'given read. Only applies to SAM/BAM input!. '
             'Default is not to filter on mapping quality.'
    )

    parser.add_argument(
        '--iterative-quality-cutoff', dest='iterative_quality_cutoff',
        type=int,
        help='MAPQ cutoff for mapped reads. Only applies when iterative '
             'mapping is enabled: if a mapped read has MAPQ below this cutoff,'
             'it will be sent to another iteration in an attempt to find a '
             'higher quality alignment. Default is 3 for BWA and 30 for Bowtie2.'
    )

    parser.add_argument(
        '--le-inward-cutoff', dest='inward_cutoff',
        type=int,
        help="Ligation error inward cutoff. "
             "Default: no ligation error filtering."
    )

    parser.add_argument(
        '--le-outward-cutoff', dest='outward_cutoff',
        type=int,
        help="Ligation error outward cutoff. "
             "Default: no ligation error filtering."
    )

    parser.add_argument(
        '--auto-le-cutoff', dest='auto_le_cutoff',
        action='store_true',
        default=False,
        help="Automatically determine ligation error cutoffs. "
             "Use with caution, this setting has a tendency to "
             "choose large cutoffs, removing many pairs close to " 
             "the diagonal."
    )

    parser.add_argument(
        '-tmp', '--work-in-tmp', dest='tmp',
        action='store_true',
        default=False,
        help="Work in temporary directory. Copies input files and " 
             "generates output files in a temporary directory. "
             "Files will be moved to their intended destination once " 
             "an analysis step finishes. "
             "Reduces network I/O if using remote file systems."
    )

    parser.add_argument(
        '--iterative', dest='iterative',
        action='store_true',
        default=False,
        help="Map reads iteratively. Can improve mappability, "
             "especially with low-quality reads. Reads are initially " 
             "trimmed to 25bp and mapped to the reference genome. "
             "If no unique mapping location is found, the read is " 
             "extended by 3bp and the process is repeated until the " 
             "full length of the read is reached or a unique mapping " 
             "location is found."
    )

    parser.add_argument(
        '--no-sam-sort', dest='sam_sort',
        action='store_false',
        default=True,
        help="Do not sort SAM/BAM files. Sorted files are required " 
             "for the pair generating step. Only omit this if you are " 
             "supplying presorted (by read name) SAM/BAM files."
    )

    parser.add_argument(
        '--restore-coverage', dest='restore_coverage',
        action='store_true',
        default=False,
        help="Restore coverage to the original total number of reads. " 
             "Otherwise matrix entries will be contact probabilities. "
             "Only available for KR matrix balancing."
    )

    parser.add_argument(
        '--split-ligation-junction', dest='split_ligation_junction',
        action='store_true',
        default=False,
        help="Split reads at predicted ligation junction before mapping. "
             "Requires the -r argument."
    )

    parser.add_argument(
        '--no-filter-pairs', dest='filter_pairs',
        action='store_false',
        default=True,
        help="Do not filter read pairs. By default, the following "
             "filters are applied: self-ligations, PCR duplicates,"
             "restriction distance (>10kb)"
    )

    parser.add_argument(
        '--no-hic', dest='process_hic',
        action='store_false',
        default=True,
        help="Do not process pairs into Hi-C maps "
             "(stop after read pairing step)."
    )

    parser.add_argument(
        '--run-with', dest='run_with',
        default='parallel',
        help="Choose how to run the commands in fanc auto. Options: "
             "'parallel' (default): Run fanc commands on local machine, "
             "use multiprocessing parallelisation. "
             "'sge': Submit fanc commands to a Sun/Oracle Grid Engine cluster. "
             "'slurm': Submit fanc commands to a Slurm cluster. "
             "'test': Do not run fanc commands but print all commands " 
             "and their dependencies to stdout for review."
    )

    parser.add_argument(
        '--job-prefix', dest='job_prefix',
        help="Job Prefix for SGE and Slurm. "
             "Works with '--run-with sge' and --run-with slurm. "
             "Default: 'fanc_<6 random letters>_'"
    )

    parser.add_argument(
        '--grid-startup-commands', dest='grid_startup_commands',
        help="Path to a file with BASH commands that are executed "
             "before every FAN-C command that is run on a grid engine / cluster. "
             "This could, for example, include environment-specific settings, "
             "such as activation of a Python virtualenv."
    )

    parser.add_argument(
        '--grid-cleanup-commands', dest='grid_cleanup_commands',
        help="Path to a file with BASH commands that are executed "
             "after every FAN-C command that is run on a grid engine / cluster. "
             "Use this to clean the file system or environment set up with "
             "--grid-startup-commands."
    )

    parser.add_argument(
        '-f', '--force-overwrite', dest='force_overwrite',
        action='store_true',
        default=False,
        help="Force overwriting of existing files. Otherwise you will " 
             "be prompted before files are overwritten."
    )

    return parser


def file_type(file_name):
    # try to see if this is a valid pairs file
    from fanc.pairs import HicProPairGenerator, FourDNucleomePairGenerator
    try:
        _ = FourDNucleomePairGenerator(file_name)
        return 'pairs_txt'
    except ValueError:
        pass

    try:
        _ = HicProPairGenerator(file_name)
        return 'pairs_txt'
    except ValueError:
        pass

    base, extension = os.path.splitext(file_name)
    if extension in ['.sam', '.bam']:
        return 'sam'
    if extension in ['.pairs']:
        return 'pairs'
    if extension in ['.hic']:
        return 'hic'
    if extension in ['.gz', '.gzip']:
        base, extension = os.path.splitext(base)
    if extension in ['.fq', '.fastq']:
        return 'fastq'
    return None


def file_basename(file_name):
    basename = os.path.basename(os.path.splitext(file_name)[0])
    if file_name.endswith('.gz') or file_name.endswith('.gzip'):
        basename = os.path.splitext(basename)[0]
    return basename


def auto(argv, **kwargs):
    from fanc.tools.general import which

    parser = auto_parser()
    args = parser.parse_args(argv[2:])

    verbosity = kwargs.get("verbosity", 2)
    if verbosity > 0:
        verbosity_flag = '-' + 'v' * verbosity
        fanc_base_command = ['fanc', verbosity_flag]
    else:
        fanc_base_command = ['fanc']

    log_file = kwargs.get("log_file", None)
    if log_file is not None:
        fanc_base_command += ['-l', log_file]

    bin_sizes = args.bin_sizes
    split_ligation_junction = args.split_ligation_junction
    restriction_enzyme = args.restriction_enzyme
    threads = args.threads
    genome = args.genome
    genome_index = args.genome_index
    basename = args.basename
    quality_cutoff = args.quality_cutoff
    iterative_quality_cutoff = args.iterative_quality_cutoff
    tmp = args.tmp
    mapper_parallel = args.mapper_parallel
    split_fastq = args.split_fastq
    memory_map = args.memory_map
    iterative = args.iterative
    step_size = args.step_size
    sam_sort = args.sam_sort
    filter_pairs = args.filter_pairs
    inward_cutoff = args.inward_cutoff
    outward_cutoff = args.outward_cutoff
    auto_le_cutoff = args.auto_le_cutoff
    process_hic = args.process_hic
    ice = args.ice
    restore_coverage = args.restore_coverage
    run_with = args.run_with
    job_prefix = args.job_prefix
    grid_startup_commands = os.path.expanduser(args.grid_startup_commands) \
        if args.grid_startup_commands is not None else None
    grid_cleanup_commands = os.path.expanduser(args.grid_cleanup_commands) \
        if args.grid_cleanup_commands is not None else None
    force_overwrite = args.force_overwrite
    output_folder = os.path.expanduser(args.output_folder)

    file_names = [os.path.expanduser(file_name) for file_name in args.input]
    file_types = [file_type(file_name) for file_name in file_names]
    file_basenames = [file_basename(file_name) for file_name in file_names]

    for file_name in file_names:
        if not os.path.exists(file_name):
            parser.error("File '{}' does not exist!".format(file_name))

    runner = None
    if run_with == 'parallel':
        runner = ParallelTaskRunner(threads)
    elif run_with == 'sge':
        from fanc.config import config
        if which(config.sge_qsub_path) is None:
            parser.error("Using SGE not possible: "
                         "Cannot find 'qsub' at path '{}'. You can change "
                         "this path using fanc config files and the "
                         "'sge_qsub_path' parameter".format(config.sge_qsub_path))
        from fanc.tools.files import mkdir
        sge_log_dir = mkdir(output_folder, 'sge_logs')
        runner = SgeTaskRunner(log_dir=sge_log_dir, task_prefix=job_prefix,
                               startup_commands_file=grid_startup_commands,
                               cleanup_commands_file=grid_cleanup_commands)
    elif run_with == 'slurm':
        from fanc.config import config
        if which(config.slurm_sbatch_path) is None:
            parser.error("Using Slurm not possible: "
                         "Cannot find 'sbatch' at path '{}'. You can change "
                         "this path using fanc config files and the "
                         "'slurm_sbatch_path' parameter".format(config.slurm_sbatch_path))
        from fanc.tools.files import mkdir
        slurm_log_dir = mkdir(output_folder, 'slurm_logs')
        runner = SlurmTaskRunner(log_dir=slurm_log_dir, task_prefix=job_prefix,
                               startup_commands_file=grid_startup_commands,
                               cleanup_commands_file=grid_cleanup_commands)
    elif run_with == 'test':
        runner = ParallelTaskRunner(threads, test=True)
    else:
        parser.error("Runner '{}' is not valid. See --run-with "
                     "parameter for options".format(run_with))

    for i in range(len(file_types)):
        if file_types[i] not in ('fastq', 'sam', 'pairs_txt', 'pairs', 'hic'):
            import fanc
            try:
                ds = fanc.load(file_names[i], mode='r')
                if isinstance(ds, fanc.Hic):
                    file_types[i] = 'hic'
                elif isinstance(ds, fanc.ReadPairs):
                    file_types[i] = 'pairs'
                else:
                    raise ValueError("Could not detect file type using fanc load.")
            except ValueError:
                parser.error("Not a valid input file type: {}".format(file_types[i]))

    if basename is None:
        if len(file_basenames) == 1:
            basename = file_basenames[0]
        else:
            basename = []
            for pair in zip(*file_basenames):
                if pair[0] == pair[1]:
                    basename.append(pair[0])
                else:
                    break
            if len(basename) == 0:
                basename = file_basenames[0]
            else:
                if basename[-1] in ['.', '_']:
                    basename = "".join(basename[:-1])
                else:
                    basename = "".join(basename)

    if not output_folder[-1] == '/':
        output_folder += '/'

    # 0. Do some sanity checks on required flags
    is_bwa = False
    is_bowtie2 = False
    if 'fastq' in file_types:
        if args.genome_index is None:
            parser.error("Must provide genome index (-i) when mapping FASTQ files!")
        else:
            check_path = os.path.expanduser(genome_index)
            if check_path.endswith('.'):
                check_path = check_path[:-1]

            is_bowtie2 = True
            for i in range(1, 5):
                if not os.path.exists(check_path + '.{}.bt2'.format(i)):
                    is_bowtie2 = False
            for i in range(1, 3):
                if not os.path.exists(check_path + '.rev.{}.bt2'.format(i)):
                    is_bowtie2 = False

            is_bwa = True
            for ending in ('amb', 'ann', 'bwt', 'pac', 'sa'):
                if not os.path.exists(check_path + '.{}'.format(ending)):
                    is_bwa = False

            if not is_bowtie2 and not is_bwa:
                parser.error("Cannot detect Bowtie2 or BWA index.")

            if is_bowtie2 and not which('bowtie2'):
                parser.error("bowtie2 must be in PATH for mapping!")

            if is_bwa and not which('bwa'):
                parser.error("bwa must be in PATH for mapping!")

    if 'fastq' in file_types or 'sam' in file_types:
        if genome is None:
            parser.error("Must provide genome (-g) to process read pair files!")

        if restriction_enzyme is None:
            from fanc.regions import genome_regions
            try:
                genome_regions(genome)
            except ValueError:
                parser.error("Must provide restriction enzyme (-r) to process read pair files!")
        else:
            res = restriction_enzyme.split(",")
            from Bio import Restriction
            for r in res:
                try:
                    getattr(Restriction, r)
                except AttributeError:
                    parser.error("Restriction enzyme string '{}' is not recognized".format(restriction_enzyme))

    logger.info("Output folder: {}".format(output_folder))
    logger.info("Input files: {}".format(", ".join(file_names)))
    logger.info("Input file types: {}".format(", ".join(file_types)))

    logger.info("Final basename: %s (you can change this with the -n option!)" % basename)

    # 1. create default folders in root directory
    if run_with != 'test':
        logger.info("Creating output folders...")
        from ..tools.general import mkdir
        mkdir(output_folder, 'fastq')
        mkdir(output_folder, 'sam')
        mkdir(output_folder, 'pairs/')
        mkdir(output_folder, 'hic/binned')
        mkdir(output_folder, 'plots/stats')

    # 2. If input files are (gzipped) FASTQ, map them iteratively first
    fastq_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'fastq':
            continue
        fastq_files.append(i)

    sam_created = [False] * len(file_names)
    mapping_tasks = []
    if len(fastq_files) > 0:
        if genome_index.endswith('.'):
            genome_index = genome_index[:-1]

        bam_files = []
        for i, ix in enumerate(fastq_files):
            bam_file = output_folder + 'sam/' + file_basenames[ix] + '.bam'
            if not force_overwrite and os.path.exists(bam_file):
                parser.error("File exists ({}), use -f to force overwriting it.".format(bam_file))
            bam_files.append(bam_file)

            mapping_command = fanc_base_command + ['map', '-m', '25',
                                                   '-s', str(step_size),
                                                   '-t', str(threads)]

            if iterative_quality_cutoff is not None:
                mapping_command += ['-q', str(iterative_quality_cutoff)]
            if tmp:
                mapping_command.append('-tmp')
            if not mapper_parallel:
                mapping_command.append('--fanc-parallel')
            if split_fastq:
                mapping_command.append('--split-fastq')
            if memory_map:
                mapping_command.append('--memory-map')
            if not iterative:
                mapping_command.append('--no-iterative')
            if split_ligation_junction:
                mapping_command.append('--restriction-enzyme')
                mapping_command.append(restriction_enzyme)

            mapping_command += [file_names[ix], genome_index, bam_file]

            mapping_task = CommandTask(mapping_command)
            runner.add_task(mapping_task, threads=threads)
            mapping_tasks.append(mapping_task)

        for ix, i in enumerate(fastq_files):
            file_names[i] = bam_files[ix]
            file_types[i] = 'sam'
            sam_created[i] = True

    if sam_sort:
        sort_threads = min(4, threads)

        sam_sort_tasks = []
        # sort SAM files
        sam_files = []
        in_place = []
        for i in range(len(file_names)):
            if file_types[i] != 'sam':
                continue
            sam_files.append(i)
            in_place.append(sam_created[i])

        if len(sam_files) > 0:
            sorted_sam_files = []
            for i, ix in enumerate(sam_files):
                sort_command = fanc_base_command + ['sort_sam', '-t', str(sort_threads),
                                                    file_names[ix]]
                if in_place[i]:
                    sorted_sam_files.append(file_names[ix])
                else:
                    sam_path, sam_extension = os.path.splitext(file_names[ix])
                    sam_basename = os.path.basename(sam_path)
                    sorted_sam_file = os.path.join(output_folder, 'sam', sam_basename + '_sort' + sam_extension)
                    if not force_overwrite and os.path.exists(sorted_sam_file):
                        parser.error("File exists ({}), use -f to force overwriting it.".format(sorted_sam_file))

                    sorted_sam_files.append(sorted_sam_file)
                    sort_command.append(sorted_sam_file)

                if tmp:
                    sort_command.append('-tmp')

                sam_sort_task = CommandTask(sort_command)
                runner.add_task(sam_sort_task, wait_for=mapping_tasks, threads=1)
                sam_sort_tasks.append(sam_sort_task)

            for ix, i in enumerate(sam_files):
                file_names[i] = sorted_sam_files[ix]
    else:
        sam_sort_tasks = mapping_tasks

    total_pairs = 0
    pairs_txt_tasks = []
    # sort SAM files
    pairs_txt_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'pairs_txt':
            continue
        pairs_txt_files.append(i)

    if len(pairs_txt_files) > 0:
        load_threads = max(int(threads / len(pairs_txt_files)), 1)

        pairs_files = []
        for ix in pairs_txt_files:
            pairs_txt_file = file_names[ix]
            pairs_file = os.path.join(output_folder, 'pairs', '{}_{}.pairs'.format(basename, total_pairs))
            total_pairs += 1
            if not force_overwrite and os.path.exists(pairs_file):
                parser.error("File exists ({}), use -f to force overwriting it.".format(pairs_file))

            pairs_files.append(pairs_file)

            pairs_command = fanc_base_command + ['pairs', '-f',
                                                 # loading
                                                 '-g', genome,
                                                 '-t', str(load_threads)]

            if restriction_enzyme is not None:
                pairs_command += ['-r', restriction_enzyme]
            if is_bwa:
                pairs_command.append('--bwa')
            if tmp:
                pairs_command.append('-tmp')
            if sam_sort:
                pairs_command.append('-S')

            pairs_command += [pairs_txt_file, pairs_file]

            pairs_task = CommandTask(pairs_command)
            runner.add_task(pairs_task, wait_for=[], threads=load_threads)
            pairs_txt_tasks.append(pairs_task)

            pairs_files.append(pairs_file)

        for ix, i in enumerate(pairs_txt_files):
            file_names[i] = pairs_files[ix]
            file_types[i] = 'pairs'

    # load pairs directly from SAM
    sam_file_pairs = []
    i = 0
    while i < len(file_names):
        if file_types[i] == 'sam':
            if not file_types[i + 1] == 'sam':
                parser.error("Cannot create SAM pairs, because {} "
                             "is missing a partner file".format(file_names[i]))
            sam_file_pairs.append((i, i + 1))
            i += 1
        i += 1

    if len(sam_file_pairs) > 0:
        sam_to_pairs_tasks = pairs_txt_tasks
        load_threads = max(int(threads/len(sam_file_pairs)), 1)

        pairs_files = []
        for i, j in sam_file_pairs:
            if len(sam_file_pairs) > 1 or total_pairs > 0:
                pairs_file = os.path.join(output_folder, 'pairs', '{}_{}.pairs'.format(basename, total_pairs))
                total_pairs += 1
            else:
                pairs_file = output_folder + 'pairs/' + basename + '.pairs'

            if not force_overwrite and os.path.exists(pairs_file):
                parser.error("File exists ({}), use -f to force overwriting it.".format(pairs_file))

            pairs_command = fanc_base_command + ['pairs', '-f',
                                                 # loading
                                                 '-g', genome,
                                                 '-t', str(load_threads),
                                                 # filtering
                                                 '-us']
            if restriction_enzyme is not None:
                pairs_command += ['-r', restriction_enzyme]
            if quality_cutoff is not None:
                pairs_command += ['-q', str(quality_cutoff)]
            if is_bwa:
                pairs_command.append('--bwa')
            if tmp:
                pairs_command.append('-tmp')
            if sam_sort:
                pairs_command.append('-S')

            pairs_command += [file_names[i], file_names[j], pairs_file]

            pairs_task = CommandTask(pairs_command)
            runner.add_task(pairs_task, wait_for=sam_sort_tasks, threads=load_threads)
            sam_to_pairs_tasks.append(pairs_task)

            pairs_files.append(pairs_file)

        for ix, sam_pair in enumerate(reversed(sam_file_pairs)):
            file_names[sam_pair[0]] = pairs_files[ix]
            del file_names[sam_pair[1]]
            file_types[sam_pair[0]] = 'pairs'
            del file_types[sam_pair[1]]
    else:
        sam_to_pairs_tasks = pairs_txt_tasks + sam_sort_tasks

    # 7. Pairs stats and filtering
    pairs_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'pairs':
            continue
        pairs_files.append(i)

    if len(pairs_files) > 0 and filter_pairs:
        pairs_tasks = []
        for ix in pairs_files:
            pair_basename = os.path.basename(os.path.splitext(file_names[ix])[0])

            pairs_stats_file = output_folder + 'plots/stats/' + pair_basename + '.pairs.stats.pdf'
            ligation_error_file = output_folder + 'plots/stats/' + pair_basename + '.pairs.ligation_error.pdf'
            re_dist_file = output_folder + 'plots/stats/' + pair_basename + '.pairs.re_dist.pdf'

            pairs_command = fanc_base_command + ['pairs',
                                                 # filtering
                                                 '-d', '10000',
                                                 '-l', '-p', '2']

            if tmp:
                pairs_command.append('-tmp')

            if inward_cutoff is not None:
                pairs_command += ['-i', str(inward_cutoff)]
                if outward_cutoff is None:
                    pairs_command += ['-o', '0']
            if outward_cutoff is not None:
                pairs_command += ['-o', str(outward_cutoff)]
                if inward_cutoff is None:
                    pairs_command += ['-i', '0']
            if inward_cutoff is None and outward_cutoff is None and auto_le_cutoff:
                pairs_command += ['--filter-ligation-auto']

            pairs_command += ['--statistics-plot', pairs_stats_file, file_names[ix]]
            pairs_task = CommandTask(pairs_command)
            runner.add_task(pairs_task, wait_for=sam_to_pairs_tasks, threads=1)
            pairs_tasks.append(pairs_task)

            ligation_error_command = fanc_base_command + ['pairs', '--ligation-error-plot',
                                                          ligation_error_file, file_names[ix]]
            ligation_error_task = CommandTask(ligation_error_command)
            runner.add_task(ligation_error_task, wait_for=pairs_task, threads=1)

            re_dist_command = fanc_base_command + ['pairs', '--re-dist-plot',
                                                   re_dist_file, file_names[ix]]
            re_dist_task = CommandTask(re_dist_command)
            runner.add_task(re_dist_task, wait_for=pairs_task, threads=1)
    else:
        pairs_tasks = sam_to_pairs_tasks

    # 8. Pairs to Hic
    pairs_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'pairs':
            continue
        pairs_files.append(i)

    if len(pairs_files) > 0 and process_hic:
        hic_tasks = []
        hic_files = []
        for ix in pairs_files:
            hic_basename = os.path.basename(os.path.splitext(file_names[ix])[0])
            if hic_basename.endswith('_filtered'):
                hic_basename = hic_basename[:-9]
            hic_file = output_folder + 'hic/' + hic_basename + '.hic'

            if not force_overwrite and os.path.exists(hic_file):
                parser.error("File exists ({}), use -f to force overwriting it.".format(hic_file))

            hic_command = fanc_base_command + ['hic', '-f']
            if tmp:
                hic_command.append('-tmp')
            hic_command += [file_names[ix], hic_file]
            hic_task = CommandTask(hic_command)
            runner.add_task(hic_task, wait_for=pairs_tasks, threads=1)
            hic_tasks.append(hic_task)
            hic_files.append(hic_file)

        for ix, i in enumerate(pairs_files):
            file_names[i] = hic_files[ix]
            file_types[i] = 'hic'
    else:
        hic_tasks = pairs_tasks

    # 9. Merge Hic
    hic_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'hic':
            continue
        hic_files.append(i)

    if len(hic_files) > 1:
        merge_hic_tasks = []
        output_hic = output_folder + 'hic/' + basename + '.hic'

        if not force_overwrite and os.path.exists(output_hic):
            parser.error("File exists ({}), use -f to force overwriting it.".format(output_hic))

        merge_hic_command = fanc_base_command + ['hic', '-f']
        if tmp:
            merge_hic_command.append('-tmp')

        hics = [file_names[i] for i in hic_files]
        merge_hic_command += hics + [output_hic]
        merge_hic_task = CommandTask(merge_hic_command)
        runner.add_task(merge_hic_task, wait_for=hic_tasks, threads=1)
        merge_hic_tasks.append(merge_hic_task)

        file_names[hic_files[0]] = output_hic
        hic_files.pop(0)
        for ix, i in enumerate(reversed(hic_files)):
            del file_names[i]
            del file_types[i]
    else:
        merge_hic_tasks = hic_tasks

    from fanc.tools.general import human_format, str_to_int

    hic_files = []
    for i in range(len(file_names)):
        if file_types[i] != 'hic':
            continue
        hic_files.append(i)

    if len(hic_files) > 0:
        for ix in hic_files:
            hic_file = file_names[ix]
            binned_hic_file_base = output_folder + 'hic/binned/' + basename + '_'
            bin_threads = min(4, threads)

            for bin_size in bin_sizes:
                bin_size = str_to_int(str(bin_size))
                bin_size_str = human_format(bin_size, 0).lower() + 'b'

                binned_hic_file = binned_hic_file_base + bin_size_str + '.hic'

                if not force_overwrite and os.path.exists(binned_hic_file):
                    parser.error("File exists ({}), use -f to force overwriting it.".format(binned_hic_file))

                hic_basename = os.path.basename(os.path.splitext(binned_hic_file)[0])
                hic_stats_file = output_folder + 'plots/stats/' + \
                                 hic_basename + '.stats.pdf'

                hic_command = fanc_base_command + ['hic', '-f', '-b', str(bin_size),
                                                   '-r', '0.1', '-t', str(bin_threads),
                                                   '--statistics-plot', hic_stats_file]
                if tmp:
                    hic_command.append('-tmp')
                if ice:
                    hic_command.append('-i')
                else:
                    hic_command.append('-k')
                if restore_coverage:
                    hic_command.append('-c')

                hic_command += [hic_file, binned_hic_file]

                hic_task = CommandTask(hic_command)
                runner.add_task(hic_task, wait_for=merge_hic_tasks, threads=bin_threads)

    runner.run()
    return 0



"""
Module for handling genomic regions such as chromosomes, bins, and restriction fragments.

The class :class:`~RegionsTable` is an implementation of the :class:`~RegionBased`
interface from the :mod:`genomic_regions` package. More details on how to use the
:class:`~genomic_regions.RegionBased` interface can be found in the :mod:`genomic_regions`
documentation, but here is an example to get you started:

.. code::

    import fanc
    rt = fanc.RegionsTable()

    # demo how to add regions
    regions = []
    for chromosome in ['chr1', 'chr2']:
        for start in range(1, 10000, 1000):
            r = fanc.GenomicRegion(chromosome=chromosome, start=start, end=start+999)
            regions.append(r)
    rt.add_regions(regions)

    # query regions on chromosome 1
    for r in rt.regions('chr1'):
        print(r)  # chr1:1-1000, chr1:1001-2000, ..., chr1:9001-10000

The class :class:`~Chromosome` holds chromosome information, specifically the
chromosome name in the reference genome, its length in base pairs, and its DNA
sequence. Multiple :class:`~Chromosome` objects can be grouped into a :class:`~Genome`,
which provides convenient access to its chromosomes and has useful functions for
in silico digestion and genome binning.

.. code::

    import fanc

    # create chromosomes
    chromosome1 = fanc.Chromosome(name='chr1', sequence='AAGTCCGTGCTGTCGATCATAGCTAGCTAGCTA')
    chromosome2 = fanc.Chromosome(name='chr2', sequence='GTGTCGATCAAATCGAAA')
    len(chromosome1)  # 33

    # create genome
    genome = fanc.Genome()
    genome.add_chromosome(chromosome1)
    genome.add_chromosome(chromosome2)

    # in silico digestion
    restriction_fragments = genome.get_regions('MboI')  # cuts 'GATC'
    for r in restriction_fragments.regions:
        print(r)  # chr1:1-15, chr1:16-33, chr2:1-6, chr2:7-18

    # binning
    bins = genome.get_regions(10)  # cuts 'GATC'
    for r in bins.regions:
        print(r)  # chr1:1-10, chr1:11-20, chr1:21-30, chr1:31-33, chr2:1-10, chr2:11-18

    # load genome from file
    genome_from_file = Genome.from_string("hg19_chr18_19.fa")
    genome_from_file.chromosomes() # ['chr18', 'chr19']

"""

from __future__ import division, print_function

import os.path
import gzip

import numpy as np
import tables as t
from Bio import SeqIO, Restriction, Seq
from genomic_regions import RegionBased, GenomicRegion, load as gr_load
from .general import FileGroup
from .tools.files import is_fasta_file
from .tools.general import create_col_index, str_to_int

try:
    from itertools import izip as zip
except ImportError:
    pass

from future.utils import string_types
from builtins import object
import logging
logger = logging.getLogger(__name__)


__all__ = ['Chromosome', 'Genome', 'RegionsTable', 'LazyGenomicRegion', 'genome_regions']


def genome_regions(re_or_file, restriction_enzyme=None):
    """
    Obtain RE fragments or bins of equal size from a reference genome.

    :param re_or_file: Path to or :class:`~RegionBased` file with regions corresponding
                       to restriction fragments, or path to FASTA file with chromosomes
    :param restriction_enzyme: If the first argument is a FASTA file, you must
                               provide the name of a restriction enzyme here
                               for in silico genome digestion. You can also provide
                               an integer, in which case the genome will be binned into
                               regions of this size in bp.
    :return: :class:`~RegionsTable` of restriction fragments
    """
    if isinstance(re_or_file, RegionBased):
        return re_or_file

    logger.info("Getting regions")
    try:
        regions = gr_load(re_or_file)
        if isinstance(regions, RegionBased):
            return regions
        else:
            raise ValueError("Not a region-based file, trying FASTA next...")
    except (ValueError, TypeError):
        if restriction_enzyme is None:
            raise ValueError("Cannot calculate Hi-C regions. You must supply "
                             "either a restriction enzyme name or bin size")
        genome = Genome.from_string(re_or_file)
        regions = genome.get_regions(restriction_enzyme)
        genome.close()

    return regions


class Chromosome(object):
    """
    Chromosome data type.

    .. attribute:: name

        Name of the chromosome

    .. attribute:: length

        Length of the chromosome in base-pairs

    .. attribute:: sequence

        Base-pair sequence of DNA in the chromosome
    """

    def __init__(self, name=None, length=None, sequence=None):
        """
        Initialize chromosome

        :param name: Name of the chromosome
        :param length: Length of the chromosome in base-pairs
        :param sequence: Base-pair sequence of DNA in the chromosome
        """
        self.name = name.decode() if isinstance(name, bytes) else name
        self.length = length
        self.sequence = sequence.decode() if isinstance(sequence, bytes) else sequence
        if length is None and sequence is not None:
            self.length = len(sequence)
        if sequence is None and length is not None:
            self.length = length

    def __repr__(self):
        return "Name: %s\nLength: %d\nSequence: %s" % (self.name if self.name else '',
                                                       self.length if self.length else -1,
                                                       self.sequence[:20] + "..." if self.sequence else '')

    def __len__(self):
        """
        Get length of the chromosome.
        """
        return self.length

    def __getitem__(self, key):
        """
        Get object attributes by name
        """
        if key == 'name':
            return self.name
        if key == 'length':
            return self.length
        if key == 'sequence':
            return self.sequence

    @classmethod
    def from_fasta(cls, file_name, name=None, include_sequence=True):
        """
        Create a :class:`~Chromosome` from a FASTA file.

        This class method will load a FASTA file and convert it into
        a :class:`~Chromosome` object. If the FASTA file contains multiple
        sequences, the output will be a list of :class:`~Chromosome` objects.

        :param file_name: Path to the FASTA file
        :param name: Chromosome name. If None (default), will be read
                     from the FASTA file header
        :param include_sequence: If True (default), stores the chromosome
                                 sequence in memory. Else, the sequence
                                 attribute will be set to None.
        :return: :class:`~Chromosome` if there is only a single FASTA
                 sequence in the file, list(:class:`~Chromosome`) if
                 there are multiple sequences.
        """
        open_ = gzip.open if file_name.endswith('.gz') or file_name.endswith('.gzip') else open

        with open_(file_name, 'rt') as fasta_file:
            fastas = SeqIO.parse(fasta_file, 'fasta')

            chromosomes = []
            for fasta in fastas:
                if include_sequence:
                    chromosome = cls(name if name else fasta.id, length=len(fasta),
                                     sequence=str(fasta.seq))
                else:
                    chromosome = cls(name if name else fasta.id, length=len(fasta))
                chromosomes.append(chromosome)

        if len(chromosomes) == 0:
            raise ValueError("File {} does not appear to be a FASTA file".format(file_name))
        if len(chromosomes) == 1:
            return chromosomes[0]
        return chromosomes

    def get_restriction_sites(self, restriction_enzyme):
        """
        Find the restriction sites of a provided enzyme in this chromosome.

        Internally uses Biopython to find RE sites.

        :param restriction_enzyme: The name of the restriction enzyme
                                   (e.g. HindIII)
        :return: List of RE sites in base-pairs (1-based)
        """
        logger.debug("Calculating RE sites")
        try:
            re = getattr(Restriction, restriction_enzyme)
        except SyntaxError:
            raise ValueError("restriction_enzyme must be a string")
        except AttributeError:
            raise ValueError("restriction_enzyme string is not recognized: %s" % restriction_enzyme)

        return re.search(Seq.Seq(self.sequence))


class LazyGenomicRegion(GenomicRegion):
    """
    A :class:`~GenomicRegion` object with lazy attribute loading.

    This class is central to an efficient retrieval of regions from
    objects subclassing :class:`~RegionsTable`. Its handling should
    be mostly identical to :class:`~GenomicRegion`, but attributes
    will only be loaded on demand. Changes to attributes will change
    the underlying row in the HDF5 regions table if the auto_update
    parameter is set to True (default). Else you can manually call
    :func:`~LazyGenomicRegion.update`.
    """
    def __init__(self, row, ix=None, auto_update=True):
        """
        Initialise this LazyGenomicRegion.

        :param row: Pytables row from a :class:`~RegionsTable`
        :param ix: (optional) region index. Overrides "ix" set in row.
        :param auto_update: Write changed attribute data to underlying table
                            if True (default). If False, call
                            :func:`~LazyGenomicRegion.update` manually after
                            making changes.
        """
        self._row = row
        self._static_ix = ix
        self._auto_update = auto_update

    def __getattr__(self, item):
        try:
            value = self._row[item]
            value = value.decode() if isinstance(value, bytes) else value
            return value
        except KeyError:
            raise AttributeError("No such attribute: {}".format(item))

    def __setattr__(self, key, value):
        if not key.startswith('_'):
            self._row[key] = value
            if self._auto_update:
                self._row.update()
        else:
            object.__setattr__(self, key, value)

    def update(self):
        self._row.update()

    @property
    def strand(self):
        """
        Strand this region is located on as int.

        :return: 1: forward strand, -1 reverse strand, 0 or
                 None: unknown.
        """
        try:
            return self._row["strand"]
        except KeyError:
            return None

    @property
    def ix(self):
        """
        Region index.

        Location in underlying list of regions.
        """
        if self._static_ix is None:
            return self._row["ix"]
        return self._static_ix

    @property
    def attributes(self):
        """
        List of all attributes in this region.
        """
        return self._row.table.colnames


class RegionBasedWithBins(RegionBased):
    """
    Extension of :class:`~RegionBased` with support for genomic bins.

    Provides a few convenience functions for dealing with genomic bins.
    """
    def __init__(self):
        super(RegionBasedWithBins, self).__init__()

    @property
    def chromosome_bins(self):
        """
        Returns a dictionary of chromosomes and the start
        and end index of the bins they cover.

        Returned list is range-compatible, i.e. chromosome
        bins [0,5] cover chromosomes 1, 2, 3, and 4, not 5.
        """
        return self._chromosome_bins()

    def _chromosome_bins(self, *args, **kwargs):
        chr_bins = {}
        for r in self.regions(*args, **kwargs):
            if chr_bins.get(r.chromosome) is None:
                chr_bins[r.chromosome] = [r.ix, r.ix + 1]
            else:
                chr_bins[r.chromosome][1] = r.ix + 1
        return chr_bins

    @property
    def bin_size(self):
        """
        Return the length of the first region in the dataset.

        Assumes all bins have equal size.

        :return: int
        """
        return len(self.regions[0]) + 1

    def distance_to_bins(self, distance):
        """
        Convert base pairs to fraction of bins.

        :param distance: distance in base pairs
        :return: float, distance as fraction of bin size
        """
        bin_size = self.bin_size
        bin_distance = int(distance / bin_size)
        if distance % bin_size > 0:
            bin_distance += 1
        return bin_distance

    def bins_to_distance(self, bins):
        """
        Convert fraction of bins to base pairs

        :param bins: float, fraction of bins
        :return: int, base pairs
        """
        return int(self.bin_size * bins)

    def region_bins(self, *args, **kwargs):
        """
        Return slice of start and end indices spanned by a region.

        :param args: provide a :class:`~GenomicRegion` here to get
                     the slice of start and end bins of onlythis region.
                     To get the slice over all regions leave this blank.
        :return:
        """
        start_ix = None
        end_ix = None
        for i, r in enumerate(self.regions(*args, **kwargs)):
            ix = getattr(r, 'ix', i)
            if start_ix is None:
                start_ix = ix
            end_ix = ix + 1
        return slice(start_ix, end_ix, 1)


class RegionsTable(RegionBasedWithBins, FileGroup):
    """
    PyTables Table wrapper for storing genomic regions.

    This class is inherited by objects working with lists of genomic
    regions, such as equidistant bins along chromosomes in a genome
    (:class:`~fanc.hic.Hic`) or restriction fragments of genomic DNA
    (:class:`~fanc.pairs.ReadPairs`)

    Internally, each genomic region is encoded in a PyTables Table and
    the following region attributes are represented as table columns:
    ix, chromosome, start, end, and strand. To add additional region
    attributes, such as a score, use the "additional_fields" parameter
    of the __init__ method. This must be a dict where the keys are str
    and values are PyTables column descriptors, such as
    :class:`~tables.StringCol`. Example for adding a score field:

    .. code::

        import fanc
        import tables
        rt = fanc.RegionsTable(
                additional_fields={'score': tables.Float32Col()}
             )

    """

    _classid = 'REGIONSTABLE'

    class RegionDescription(t.IsDescription):
        """
        Description of a genomic region for PyTables Table
        """
        ix = t.Int32Col(pos=0)
        chromosome = t.StringCol(100, pos=1)
        start = t.Int64Col(pos=2)
        end = t.Int64Col(pos=3)
        strand = t.Int8Col(pos=4)
        _mask_ix = t.Int32Col(pos=5)

    class ChromosomeDescription(t.IsDescription):
        """
        Description of the chromosomes in this object.
        """
        ix = t.Int32Col(pos=0)
        name = t.StringCol(100, pos=1)
        start_bin = t.Int32Col(pos=2)
        end_bin = t.Int32Col(pos=3)
        size = t.Int32Col(pos=4)

    def __init__(self, file_name=None, mode='a', tmpdir=None,
                 additional_fields=None, _table_name_regions='regions'):
        """
        Initialize region table.

        :param file_name: Path to file or None for in-memory file
        :param mode: File mode. Defaults to 'a' (append). Use 'r' for read-only
                     access, and 'w' for write mode that will overwrite any
                     previous file content.
        :param tmpdir: If True, will copy an existing or create a new file to a
                       temporary directory. You can also pass the path to a folder
                       here directly. The file is copied to the location given by
                       file_name when calling :func:`~RegionsTable.close`
        :param additional_fields: Dictionary of additional columns to be appended
                                  to the PyTables table holding the genomic regions.
                                  By default, the columns are: ix, chromosome, start,
                                  end, strand, and _mask_ix. This must be a dict
                                  where the keys are str and values are PyTables
                                  column descriptors
        :param _table_name_regions: (Internal) name of the HDF5
                                    node that stores data for this
                                    object
        """
        self._regions_dirty = False

        file_exists = False
        if file_name is not None and os.path.exists(os.path.expanduser(file_name)):
            file_exists = True

        FileGroup.__init__(self, _table_name_regions, file_name, mode=mode, tmpdir=tmpdir)

        if file_exists and mode != 'w':
            self._regions = self._group.regions
            try:
                self._chromosomes_info = self._group.chromosomes
            except t.NoSuchNodeError:
                self._chromosomes_info = None
        else:
            basic_fields = dict()
            hidden_fields = dict()
            for field, description in RegionsTable.RegionDescription().columns.copy().items():
                if field.startswith('_'):
                    hidden_fields[field] = description
                else:
                    basic_fields[field] = description

            current = len(basic_fields)
            if additional_fields is not None:
                if (not isinstance(additional_fields, dict) and
                        issubclass(additional_fields, t.IsDescription)):
                    # IsDescription subclass case
                    additional_fields = additional_fields.columns

                # add additional user-defined fields
                for key, value in sorted(additional_fields.items(),
                                         key=lambda x: x[1]._v_pos if x[1]._v_pos is not None else 1):
                    if key not in basic_fields:
                        value._v_pos = current
                        current += 1
                        basic_fields[key] = value
            # add hidden fields
            for key, value in sorted(hidden_fields.items(),
                                     key=lambda x: x[1]._v_pos if x[1]._v_pos is not None else 1):
                value._v_pos = current
                current += 1
                basic_fields[key] = value

            self._regions = t.Table(self._group, 'regions', basic_fields, expectedrows=1000000)
            self._chromosomes_info = t.Table(self._group, 'chromosomes',
                                             RegionsTable.ChromosomeDescription,
                                             expectedrows=100)

            # create indexes
            create_col_index(self._regions.cols.ix)
            create_col_index(self._regions.cols.start)
            create_col_index(self._regions.cols.end)

        self._max_region_ix = None

    def _update_chromosomes_info(self):
        try:
            self._chromosomes_info = self._group.chromosomes
        except t.NoSuchNodeError:
            try:
                self._chromosomes_info = t.Table(self._group, 'chromosomes',
                                                 RegionsTable.ChromosomeDescription,
                                                 expectedrows=100)
            except (t.FileModeError, t.HDF5ExtError):
                self._chromosomes_info = None

        if self._chromosomes_info is not None:
            try:
                self._chromosomes_info.remove_rows(0)
                self._chromosomes_info.flush()
            except t.HDF5ExtError:
                logger.error("File not open for writing, cannot update chromosome table!")
                return

            chromosomes_info = []
            for row in self._regions.iterrows():
                if len(chromosomes_info) == 0 or row['chromosome'] != chromosomes_info[-1][1]:
                    chromosomes_info.append([len(chromosomes_info), row['chromosome'],
                                             row['ix'], row['ix'], row['end']])
                else:
                    chromosomes_info[-1][3] = row['ix']
                    chromosomes_info[-1][4] = row['end']

            row = self._chromosomes_info.row
            for info in chromosomes_info:
                row['ix'] = info[0]
                row['name'] = info[1]
                row['start_bin'] = info[2]
                row['end_bin'] = info[3]
                row['size'] = info[4]
                row.append()
            self._chromosomes_info.flush()

    def _flush_regions(self):
        """
        Write buffered regions to PyTables Table.
        """
        if self._regions_dirty:
            self._regions.flush()
            self._update_chromosomes_info()
            self._regions_dirty = False

    def flush(self):
        """
        Write buffered data to file.
        """
        self._flush_regions()

    def _add_region(self, region, preserve_attributes=True, *args, **kwargs):
        """
        Basic function to add a region to this object.

        :param region: :class:`~GenomicRegion`
        :param preserve_attributes: If True, will attempt to copy all region
                                    attributes to this region table. If False,
                                    will only use chromosome, start, end, and
                                    strand
        :param args: Currently not used
        :param kwargs: Currently not used
        :return: Region index of the added region.
        """
        self._regions_dirty = True
        if self._max_region_ix is None:
            self._max_region_ix = len(self._regions) - 1
        ix = self._max_region_ix + 1

        # actually append
        row = self._regions.row
        row['ix'] = ix
        row['chromosome'] = region.chromosome
        row['start'] = region.start
        row['end'] = region.end
        if hasattr(region, 'strand') and region.strand is not None:
            row['strand'] = region.strand

        if preserve_attributes:
            for name in self._regions.colnames[5:]:
                if hasattr(region, name):
                    row[name] = getattr(region, name)

        row.append()

        # if ix > getattr(self.meta, 'max_region_ix', -1):
        #     self.meta['max_region_ix'] = ix
        self._max_region_ix += 1

        return ix

    def chromosomes(self):
        """
        List all chromosomes in this regions table.
        :return: list of chromosome names.
        """
        if self._chromosomes_info is not None:
            return [name.decode() for name in self._chromosomes_info.col("name")]

        chromosomes_set = set()
        chromosomes = []
        for region in self.regions(lazy=True):
            if region.chromosome not in chromosomes_set:
                chromosomes_set.add(region.chromosome)
                chromosomes.append(region.chromosome)
        return chromosomes

    @property
    def chromosome_lengths(self):
        if self._chromosomes_info is not None:
            cr = {}
            for row in self._chromosomes_info.iterrows():
                cr[row['name'].decode()] = row['size']
            return cr
        else:
            return super(RegionsTable, self).chromosome_lengths

    def _chromosome_bins(self, *args, **kwargs):
        if self._chromosomes_info is not None:
            cb = {}
            for row in self._chromosomes_info.iterrows():
                cb[row['name'].decode()] = [row['start_bin'], row['end_bin'] + 1]
            return cb
        else:
            return RegionBasedWithBins._chromosome_bins(self, *args, **kwargs)

    def add_regions(self, regions, *args, **kwargs):
        """
        Bulk insert multiple genomic regions.

        :param regions: List (or any iterator) with objects that
                        describe a genomic region. See
                        :class:`~RegionsTable.add_region` for options.
        """
        self._regions_dirty = True
        for i, region in enumerate(regions):
            self.add_region(region, *args, **kwargs)

        self._flush_regions()

    def region_data(self, key, value=None):
        """
        Retrieve or add vector-data to this object. If there is existing data in this
        object with the same name, it will be replaced

        :param key: Name of the data column
        :param value: vector with region-based data (one entry per region)
        """
        if key not in self._regions.colnames:
            raise KeyError("{} is unknown region attribute".format(key))

        if value is not None:
            for i, row in enumerate(self._regions):
                row[key] = value[i]
                row.update()
            self._flush_regions()

        return (row[key] for row in self._regions)

    def _get_region_ix(self, region):
        """
        Get index from other region properties (chromosome, start, end)
        """
        condition = "(start == %d) & (end == %d) & (chromosome == b'%s')"
        condition %= region.start, region.end, region.chromosome
        for res in self._regions.where(condition):
            return res["ix"]
        return None

    def _row_to_region(self, row, lazy_region=None):
        """
        Convert a PyTables row to :class:`~GenomicRegion`.

        :param row: PyTables row object
        :param lazy_region: (optional) :class:`~LazyGenomicRegion` that is
                            used for loading attributes.
        :return: :class:`~GenomicRegion` or :class:`~LazyGenomicRegion`
        """
        if lazy_region is not None:
            lazy_region._row = row
            return lazy_region

        kwargs = {}
        for name in self._regions.colnames:
            if name not in RegionsTable.RegionDescription().columns.keys():
                value = row[name]
                value = value.decode() if isinstance(value, bytes) else value
                kwargs[name] = value

        try:
            mask_ix = row['_mask_ix']
        except (KeyError, ValueError):
            mask_ix = 0

        return GenomicRegion(chromosome=row["chromosome"].decode(), start=row["start"],
                             end=row["end"], ix=row["ix"], _mask_ix=mask_ix, **kwargs)

    def _region_iter(self, lazy=False, auto_update=True, *args, **kwargs):
        """
        Iterate over all genomic regions.
        """
        if lazy:
            lazy_region = LazyGenomicRegion(row=None, auto_update=auto_update)
        else:
            lazy_region = None

        for row in self._regions:
            yield self._row_to_region(row, lazy_region=lazy_region)

    def _region_subset(self, region, lazy=False, auto_update=True, *args, **kwargs):
        """
        Iterate over a range of genomic regions.
        """
        if lazy:
            lazy_region = LazyGenomicRegion(row=None, auto_update=auto_update)
        else:
            lazy_region = None

        for row in self._subset_rows(region):
            sub_region = self._row_to_region(row, lazy_region=lazy_region)
            yield sub_region

    def _get_regions(self, key, *args, **kwargs):
        """
        Get specific regions by key.
        """
        res = self._regions[key]

        if isinstance(res, np.ndarray):
            regions = []
            for region in res:
                regions.append(self._row_to_region(region))
            return regions
        else:
            return self._row_to_region(res)

    def _region_len(self):
        """
        Get the number of regions in this object.
        """
        return len(self._regions)

    def _subset_rows(self, key):
        """
        Iterate over a subset of regions given the specified key.

        :param key: A :class:`~GenomicRegion` object,
                    or a list of the former. Also accepts slices and integers
        :return: Iterator over the specified subset of regions
        """
        if isinstance(key, slice):
            for row in self._regions.where("(ix >= {}) & (ix < {})".format(key.start, key.stop)):
                yield row
        elif isinstance(key, int):
            yield self._regions[key]
        elif isinstance(key, list) and len(key) > 0 and isinstance(key[0], int):
            for ix in key:
                yield self._regions[ix]
        else:
            if isinstance(key, string_types):
                key = GenomicRegion.from_string(key)

            if isinstance(key, GenomicRegion):
                keys = [key]
            else:
                keys = key

            for k in keys:
                if isinstance(k, string_types):
                    k = GenomicRegion.from_string(k)

                query = '('
                if k.chromosome is not None:
                    query += "(chromosome == b'%s') & " % k.chromosome
                if k.end is not None:
                    query += "(start <= %d) & " % k.end
                if k.start is not None:
                    query += "(end >= %d) & " % k.start
                if query.endswith(' & '):
                    query = query[:-3]
                query += ')'

                if len(query) == 2:
                    for row in self._regions:
                        yield row
                else:
                    for row in self._regions.where(query):
                        yield row


class Genome(FileGroup):
    """
    Class representing a collection of chromosomes.

    Provides some convenience batch
    methods that call :class:`~Chromosome` methods for every
    chromosome in this object.
    """

    _classid = 'GENOME'

    class ChromosomeDefinition(t.IsDescription):
        name = t.StringCol(255, pos=0)
        length = t.Int64Col(pos=1)

    def __init__(self, file_name=None, chromosomes=None, mode='a', tmpdir=None,
                 _table_name_chromosomes='chromosomes'):
        """
        Build :class:`~Genome` from a list of chromosomes or load
        previously saved object.

        :param file_name: Path of the file to load or to save to.
        :param chromosomes: List of chromosomes to load into this
                            object.
        """
        FileGroup.__init__(self, _table_name_chromosomes, file_name, mode=mode, tmpdir=tmpdir)

        # check if this is an existing regions file
        try:
            self._sequences = self._group.sequences
        except t.NoSuchNodeError:
            self._sequences = self.file.create_vlarray(self._group, 'sequences', t.VLStringAtom())

        try:
            self._chromosome_table = self._group.chromosomes
        except t.NoSuchNodeError:
            try:
                self._chromosome_table = self.file.create_table(self._group, 'chromosomes',
                                                                Genome.ChromosomeDefinition)
            except t.FileModeError:
                self._chromosome_table = None
                pass

        if chromosomes is not None:
            if isinstance(chromosomes, Chromosome):
                chromosomes = [chromosomes]

            for chromosome in chromosomes:
                self.add_chromosome(chromosome)

    def chromosomes(self):
        """
        Get list of chromosomes in this object
        """
        return self._names

    @property
    def _names(self):
        if self._chromosome_table is None:
            try:
                return self.meta['chromosome_names']
            except KeyError:
                return []
        else:
            return [row['name'].decode() if isinstance(row['name'], bytes) else row['name']
                    for row in self._chromosome_table]

    @_names.setter
    def _names(self, names):
        if self._chromosome_table is None:
            self.meta['chromosome_names'] = names
        else:
            counter = 0
            for i, row in enumerate(self._chromosome_table):
                row['name'] = names[i]
                row.update()
                counter += 1
            for i in range(counter, len(names)):
                self._chromosome_table.row['name'] = names[i]
                self._chromosome_table.row.append()
        self._chromosome_table.flush()

    @property
    def _lengths(self):
        if self._chromosome_table is None:
            try:
                return self.meta['chromosome_lengths']
            except KeyError:
                return []
        else:
            return [row['length'] for row in self._chromosome_table]

    @_lengths.setter
    def _lengths(self, lengths):
        if self._chromosome_table is None:
            self.meta['chromosome_lengths'] = lengths
        else:
            counter = 0
            for i, row in enumerate(self._chromosome_table):
                row['length'] = lengths[i]
                row.update()
                counter += 1
            for i in range(counter, len(lengths)):
                self._chromosome_table.row['length'] = lengths[i]
                self._chromosome_table.row.append()
        self._chromosome_table.flush()
        self.meta['chromosome_lengths'] = lengths

    @classmethod
    def from_folder(cls, folder_name, file_name=None, exclude=None,
                    include_sequence=True, tmpdir=None):
        """
        Load every FASTA file from a folder as a chromosome.

        :param folder_name: Path to the folder to load
        :param file_name: File to save Genome object to
        :param exclude: List or set of chromosome names that
                        should NOT be loaded
        :param include_sequence: If True, will save the
                                 chromosome sequences in the
                                 Genome object
        """
        chromosomes = []
        folder_name = os.path.expanduser(folder_name)
        for f in os.listdir(folder_name):
            try:
                chromosome = Chromosome.from_fasta(folder_name + "/" + f,
                                                   include_sequence=include_sequence)
                logger.info("Adding chromosome %s" % chromosome.name)
                if exclude is None:
                    chromosomes.append(chromosome)
                elif chromosome.name not in exclude:
                    chromosomes.append(chromosome)
            except (ValueError, IOError):
                pass

        return cls(chromosomes=chromosomes, file_name=file_name, tmpdir=tmpdir)

    @classmethod
    def from_string(cls, genome_string, file_name=None, tmpdir=None, mode='a'):
        """
        Convenience function to load a :class:`~Genome` from a string.

        :param genome_string: Path to FASTA file, path to folder with
                              FASTA files, comma-separated list of
                              paths to FASTA files, path to HDF5 file
        :param file_name: Path to save file
        :return: A :class:`~Genome` object
        """
        # case 1: FASTA file = Chromosome
        if is_fasta_file(genome_string):
            chromosomes = Chromosome.from_fasta(genome_string)
            genome = cls(chromosomes=chromosomes, file_name=file_name, tmpdir=tmpdir)
        # case 2: Folder with FASTA files
        elif os.path.isdir(genome_string):
            genome = cls.from_folder(genome_string, file_name=file_name, tmpdir=tmpdir)
        # case 3: path to HDF5 file
        elif os.path.isfile(genome_string):
            genome = cls(genome_string, tmpdir=tmpdir, mode=mode)
        # case 4: List of FASTA files
        else:
            chromosome_files = genome_string.split(',')
            chromosomes = []
            for chromosome_file in chromosome_files:
                chromosome = Chromosome.from_fasta(os.path.expanduser(chromosome_file))
                chromosomes.append(chromosome)
            genome = cls(chromosomes=chromosomes, file_name=file_name, tmpdir=tmpdir)

        return genome

    def __getitem__(self, key):
        """
        Get Genome table subsets.

        If the result is one or more rows, they will be converted to
        :class:`~Chromosome` objects, if the result is a column, it
        will be returned without conversion.
        """
        names = self._names
        lengths = self._lengths

        if isinstance(key, string_types):
            key = names.index(key)

        if isinstance(key, int):
            return Chromosome(name=names[key], length=lengths[key], sequence=self._sequences[key])
        elif isinstance(key, slice):
            l = []
            start = key.start if key.start is not None else 0
            stop = key.stop if key.stop is not None else len(names)
            step = key.step if key.step is None else 1

            for i in range(start, stop, step):
                c = Chromosome(name=names[i], length=lengths[i], sequence=self._sequences[i])
                l.append(c)
            return l
        else:
            l = []
            for i in key:
                if isinstance(i, string_types):
                    i = names.index(i)
                c = Chromosome(name=names[i], length=lengths[i], sequence=self._sequences[i])
                l.append(c)
            return l

    def __len__(self):
        return len(self._names)

    def __iter__(self):
        """
        Get iterator over :class:`~Chromosome` objects.
        """
        this = self

        class Iter(object):
            def __init__(self):
                self.current = 0

            def __iter__(self):
                self.current = 0
                return self

            def __next__(self):
                if self.current >= len(this):
                    raise StopIteration
                self.current += 1
                return this[self.current - 1]

        return Iter()

    def add_chromosome(self, chromosome):
        """
        Add a :class:`~Chromosome` to this object.

        Will choose suitable defaults for missing attributes.

        :param chromosome: :class:`~Chromosome` object or similar
                           object (e.g. dict) with the same fields
        """
        i = len(self._names)

        n = str(i)
        if chromosome.name is not None:
            n = chromosome.name

        l = 0
        if chromosome.length is not None:
            l = chromosome.length

        s = ''
        if chromosome.sequence is not None:
            s = chromosome.sequence
            if l == 0:
                l = len(s)

        self._chromosome_table.row['name'] = n
        self._chromosome_table.row['length'] = l
        self._chromosome_table.row.append()

        # self._names = self._names + [n]
        # self._lengths = self._lengths + [l]
        self._sequences.append(s.encode('utf-8'))
        self._sequences.flush()
        self._chromosome_table.flush()

    def get_regions(self, split, file_name=None, chromosomes=None):
        """
        Extract genomic regions from genome.

        Provides two options:

        - Splits chromosomes at restriction sites if the split
          parameter is the name of a restriction enzyme.

        - Splits chromosomes at equi-distant points if split
          is an integer

        :param split: Name of a restriction enzyme or positive
                      integer
        :param file_name: Name of a file if the result of this
                          method should be saved to file
        :param chromosomes: List of chromosome names to include. Default: all
        :return: :class:`~GenomicRegions`
        """

        regions = RegionsTable(file_name=file_name)

        if isinstance(split, string_types) or isinstance(split, int):
            splits = [split]
        else:
            splits = split

        region_list = []
        for chromosome in self:
            if chromosomes is not None and chromosome.name not in chromosomes:
                continue
            split_locations = []
            for split in splits:
                if isinstance(split, string_types) and hasattr(Restriction, split):
                    split_locations += chromosome.get_restriction_sites(split)
                elif isinstance(split, int) or isinstance(split, string_types):
                    split = str_to_int(split)
                    for i in range(split, len(chromosome) - 1, split):
                        split_locations.append(i)
                else:
                    raise ValueError("split argument {} not supported".format(split))

            split_locations = sorted(set(split_locations))

            for i in range(0, len(split_locations)):
                if i == 0:
                    region = GenomicRegion(start=1, end=int(split_locations[i]), chromosome=chromosome.name)
                else:
                    region = GenomicRegion(start=int(split_locations[i - 1] + 1),
                                           end=int(split_locations[i]), chromosome=chromosome.name)

                region_list.append(region)

            # add last node
            if len(split_locations) > 0:
                region = GenomicRegion(start=int(split_locations[len(split_locations) - 1] + 1),
                                       end=int(chromosome.length), chromosome=chromosome.name)
            else:
                region = GenomicRegion(start=1, end=int(chromosome.length), chromosome=chromosome.name)
            region_list.append(region)

        regions.add_regions(region_list)

        return regions

    def sub_sequence(self, chromosome, start=None, end=None):
        """
        Extract the chromosome DNA sequence between start and end.

        :param chromosome: Name of chromosome
        :param start: start position in bp (1-based, inclusive)
        :param end: end position in bp (1-based, inclusive)
        :return: str
        """
        if start is not None:
            selection_region = GenomicRegion(chromosome=chromosome, start=start, end=end)
        elif isinstance(chromosome, GenomicRegion):
            selection_region = chromosome
        else:
            selection_region = GenomicRegion.from_string(chromosome)

        res_chromosome = self[selection_region.chromosome]
        if selection_region.start is None:
            return res_chromosome.sequence
        return res_chromosome.sequence[selection_region.start - 1:selection_region.end]
